#ifndef MTASK_H
#define MTASK_H

#define MIN_PRIO		0
#define DEFAULT_PRIO	50
#define MAX_PRIO		(-1)
#define FOREVER			(-1)

typedef unsigned Time_t;
typedef unsigned Prio_t;

typedef enum
{ 
	false, 
	true 
} 
bool;

typedef enum 
{ 
	TaskSuspended, 
	TaskReady, 
	TaskCurrent, 
	TaskDelaying, 
	TaskWaiting, 
	TaskSending, 
	TaskReceiving, 
	TaskTerminated 
} 
TaskState_t;

typedef struct Task_t Task_t;

typedef struct
{
	char *			name;
	Task_t *		head;
	Task_t *		tail;
}
TaskQueue_t;

struct Task_t
{
	char *			name;
	TaskState_t		state;
	Prio_t			priority;
	unsigned		atomic_level;
	char *			stack;
	unsigned 		esp;
	unsigned		ss;
	void *			math_data;
	TaskQueue_t	*	queue;
	Task_t *		prev;
	Task_t *		next;
	bool			success;
	bool			in_time_q;
	Task_t *		time_prev;
	Task_t *		time_next;
	Time_t			msecs;
	void *			data;
	Task_t *		from;
	void *			msg;
	unsigned 		size;
	TaskQueue_t 	send_queue;
};

typedef void (*TaskFunc_t)(void *arg);
typedef bool (*Handler_t)(unsigned irq);
typedef void (*Callback_t)(void);
typedef void (*Switcher_t)(Task_t *save, Task_t *restore);

/* API */

void			InitMTask(Time_t mspertick, Time_t timeslice);
Task_t *		CreateTask(TaskFunc_t func, unsigned stacksize, void *arg, 
									char *name, Prio_t priority);
Task_t *		CurrentTask(void);
void			DeleteTask(Task_t *task);

void			ProtectMath(Task_t *task);
Prio_t			GetPriority(Task_t *task);
void			SetPriority(Task_t *task, Prio_t priority);
void			Suspend(Task_t *task);
void			Ready(Task_t *task);

TaskQueue_t *	CreateQueue(char *name);
void			DeleteQueue(TaskQueue_t *queue);
bool			WaitQueue(TaskQueue_t *queue);
bool			WaitQueueTimed(TaskQueue_t *queue, Time_t msecs);
bool			SignalQueue(TaskQueue_t *queue);
void			FlushQueue(TaskQueue_t *queue, bool success);

bool			Send(Task_t *to, void *msg, unsigned size);
bool			SendCond(Task_t *to, void *msg, unsigned size);
bool			SendTimed(Task_t *to, void *msg, unsigned size, Time_t msecs);
bool			Receive(Task_t **from, void *msg, unsigned *size);
bool			ReceiveCond(Task_t **from, void *msg, unsigned *size);
bool			ReceiveTimed(Task_t **from, void *msg, unsigned *size,
									Time_t msecs);
void			Pause(void);
void			Yield(void);
void			Scheduler(void);
void			Delay(Time_t msecs);
void			DelayUntil(Time_t deadline);
void			Exit(void);

void			Atomic(void);
void			Unatomic(void);

void			SetHandler(unsigned irq, Handler_t handler);
void			EndInterrupt(void);
void 			DisableIRQ(unsigned irq);
void 			EnableIRQ(unsigned irq);

void *			Malloc(unsigned size);
void *			Realloc(void *mem, unsigned size);
char *			StrDup(char *str);
void 			Free(void *mem);

Time_t			GetTime(void);
void			SetData(Task_t *task, void *data);
void			SetSwitcher(Switcher_t switcher);
void			SetTimerCallback(Callback_t callback);

void			Panic(char *msg);

void			DisableInts(void);
void			RestoreInts(void);

#endif
